from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    process_classes,
    GenericDirective,
    add_attribute_to_tag,
    create_role,
    create_directive,
    generic_visit,
    generic_depart,
    get_node,
    get_generic_translator_class,
    add_role_and_directive,
    BaseNode,
    BaseClassNode,
    add_visit_depart,
    process_style,
)

TAG = 'a'

class link_node(BaseNode): pass
class link_button_node(BaseClassNode):
    classes=['button']
class ghost_button_node(BaseClassNode):
    classes=['ghost', 'button']

def setup_link(app):
    app.add_node(link_node)
    app.add_node(ghost_button_node)
    app.add_node(link_button_node)
    app.add_directive('link', LinkDirective)
    app.add_directive('ghost-button-link', GhostButtonLinkDirective)
    app.add_directive('button-link', ButtonLinkDirective)

class LinkDirective(GenericDirective):
    required_arguments = 1
    node_type = link_node

    def run(self):
        node = super().run()[0]
        node['target'] = self.arguments[0]
        return [node]

class GhostButtonLinkDirective(LinkDirective):
    node_type = ghost_button_node 
    
class ButtonLinkDirective(LinkDirective):
    node_type = link_button_node        

class LinkTranslator(HTMLTranslator):
    
    def visit_link_node(self, node):
        content = f"<{TAG} href=\"{node['target']}\">"
        content = process_classes(content, node, TAG)
        content = process_style(content, node, TAG)
        self.body.append(content)

    def depart_link_node(self, node):
        generic_depart(self, TAG)

    def visit_ghost_button_node(self, node):
        self.visit_link_node(node)

    def depart_ghost_button_node(self, node):
        generic_depart(self, TAG)
        
    def visit_link_button_node(self, node):
        self.visit_link_node(node)

    def depart_link_button_node(self, node):
        generic_depart(self, TAG)



from copy import deepcopy

from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    BaseNode,
    GenericDirective,
    add_visit_depart
)

class web_node(BaseNode): pass
class iframe_node(BaseNode): pass
class figcaption_node(BaseNode): pass

# tag to node
web_map = {
    'figure': web_node,
    'iframe': iframe_node,
    'figcaption': figcaption_node, 
}

class WebDirective(GenericDirective):
    node_type = web_node
    required_arguments = 1
    option_spec = deepcopy(GenericDirective.option_spec)
    option_spec.update({
        # wrap will be deprecated
        'height': directives.unchanged,
        'width': directives.unchanged,
    })

    def run(self):
        node = super().run()[0]
        iframe = iframe_node()
        node += iframe
        if self.content:
            figcaption = figcaption_node()
            figcaption = self._process_content(figcaption)
            node += figcaption
        height = 'height'
        if height in self.options:
            height_value = self.options[height]
        else:
            height_value = 600
        width = 'width'
        if width in self.options:
            width_value = self.options[width]
        else:
            width_value = 800
        iframe['attributes'] = {
            'src': self.arguments[0],
            'width': str(height_value),
            'height': str(width_value),
        }
        return [node]

def setup_web(app):
    for node_type in web_map.values():
        app.add_node(node_type)
    app.add_directive('web', WebDirective)

class WebTranslator(HTMLTranslator):
    pass

for tag in web_map:
    node_type = web_map[tag]
    add_visit_depart(WebTranslator, node_type.__name__, tag)


from os import path
from typing import Any, Dict, Iterator, Set, Tuple
import shutil
from pathlib import Path

from docutils.io import StringOutput
from docutils.nodes import Node
from sphinx.builders import Builder
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.util.osutil import ensuredir, os_path

from .webslides_writer import WebslidesWriter, WebslidesTranslator

class WebslidesBuilder(StandaloneHTMLBuilder):
    name = 'webslides'
    default_translator_class = WebslidesTranslator
    



from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    generic_visit,
    BaseNode,
    GenericDirective,
)

TAG = 'hr'

class topic_shift_node(BaseNode): pass

def topic_shift_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    node = topic_shift_node()
    return [node], []

class TopicShiftDirective(GenericDirective):
    node_type = topic_shift_node

def setup_hr(app):
    app.add_node(topic_shift_node)
    app.add_role('hr', topic_shift_role)
    app.add_directive('hr', TopicShiftDirective)

class HRTranslator(HTMLTranslator):

    def visit_topic_shift_node(self, node):
        generic_visit(self, TAG, node)

    def depart_topic_shift_node(self, node):
        pass


from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    generic_visit,
    generic_depart,
)

TAG = 'figure'
CAPTION_TAG = 'figcaption'

# TODO: The Figure directive will need to be replaced
# to allow complex caption contents like links and svg
# The legend functionality is also problematic

class FigureTranslator(HTMLTranslator):
    """ webslides needs a figure tag
    around the image.  No super() because
    visit_figure adds a div that ruins things"""
    
    def visit_figure(self, node):
        generic_visit(self, TAG, node)

    def depart_figure(self, node):
        generic_depart(self, TAG)

    def visit_caption(self, node):
        generic_visit(self, CAPTION_TAG, node)

    def depart_caption(self, node):
        generic_depart(self, CAPTION_TAG)


from bs4 import BeautifulSoup
from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

def add_class_to_tag(html: str, tag_name: str, class_name: str) -> str:
    return add_attribute_to_tag(html, tag_name, 'class', class_name)

def add_attribute_to_tag(html: str, tag_name: str, attribute_name: str, value: str) -> str:
    soup = BeautifulSoup(html, 'html.parser')
    # value None is for attributes that do not assign
    if value is None:
        result = None
    else:
        result = soup.find(tag_name).get(attribute_name, []) + [value]
    soup.find(tag_name)[attribute_name] = result
    # bs4 adds a close tag, removing
    return str(soup).replace(f'</{tag_name}>', '')

def process_classes(content: str, node, tag_name: str) -> str:
    for class_name in node['classes']:
        content = add_class_to_tag(
            content, tag_name, class_name
        )
    return content

def process_attributes(content: str, node, tag_name: str) -> str:
    a = 'attributes' 
    if a in node:
        for label in node[a]:
            content = add_attribute_to_tag(
                content,
                tag_name,
                label,
                node[a][label],
            )
    return content

def process_style(content: str, node, tag_name: str) -> str:
    if 'style' in node:
        if node['style'] != '':
            content = add_attribute_to_tag(
                content,
                tag_name,
                'style',
                node['style']
            )
    return content

def create_role(node_type, classes: list = None):
    def role(name, rawtext, text, lineno, inliner, options={}, content=[]):
        node = node_type()
        if classes is not None:
            node['classes'] = classes
        node += nodes.Text(text)
        return [node], []
    return role

class GenericDirective(SphinxDirective):
    option_spec = {
        'classes': directives.unchanged,
        'style': directives.unchanged,
        'horizontal-alignment': directives.unchanged,
    }
    has_content = True

    def run(self):
        node = self.node_type()
        return self.generic_process(node)

    def generic_process(self, node):
        node = self.set_horizontal_alignment(node)
        node = self.set_class_options(node)
        node = self.set_style(node)
        if self.has_content:
            node = self._process_content(node)
        return [node]    
    
    def _process_content(self, node):        
        par = nodes.paragraph()
        self.state.nested_parse(self.content, self.content_offset, par)
        node += par
        return node

    def set_style(self, node):
        if 'style' not in node:
            node['style'] = ''
        if 'style' in self.options:
            node['style'] = node['style'] + self.options['style']
        return node

    def set_class_options(self, node):
        if 'classes' not in node:
            node['classes'] = []
        if 'classes' in self.options:
            node['classes'] = node['classes'] + self.options['classes'].split(' ')
        return node

    def set_horizontal_alignment(self, node):
        ha = 'horizontal-alignment'
        if ha in self.options:
            requested_alignment = self.options[ha].lower()
            if requested_alignment == 'center':
                node.add_class('aligncenter')
            elif requested_alignment == 'left':
                node.add_class('alignleft')
            elif requested_alignment == 'right':
                node.add_class('alignright')
            else:
                print(f'ERROR: Horizontal alignment {requested_alignment} unknown, ignoring...')
        return node

    

def create_directive(requested_node_type):
    class SimpleDirective(GenericDirective):
        node_type = requested_node_type

    return SimpleDirective

def generic_visit(translator, tag: str, node):
    content = f'<{tag}>'
    content = process_classes(content, node, tag)
    content = process_style(content, node, tag)
    content = process_attributes(content, node, tag)
    translator.body.append(content)

def generic_depart(translator, tag):
    translator.body.append(f'</{tag}>')

def create_generic_visit(tag: str):
    def a_visit(self, node):
        generic_visit(self, tag, node)
    return a_visit

def create_generic_depart(tag: str):
    def a_depart(self, node):
        generic_depart(self, tag)
    return a_depart

def get_generic_translator_class(tag: str, node_types: list):
    class GenericClass(HTMLTranslator):
        pass
    for node_type in node_types:
        GenericClass = add_visit_depart(GenericClass, node_type, tag)
    return GenericClass

def add_visit_depart(GenericClass, node_type:str, tag: str):
    setattr(
        GenericClass, 
        f'visit_{node_type}', 
        create_generic_visit(tag),
    )
    setattr(
        GenericClass, 
        f'depart_{node_type}', 
        create_generic_depart(tag),
    )
    return GenericClass

def get_generic_directive(requested_node_type) -> GenericDirective:
    class AGenericDirective(GenericDirective):
        node_type = requested_node_type

    return AGenericDirective

def get_node(node_name: str, classes: list = None, input_parent_class=None):
    if input_parent_class is None:
        ParentClass = nodes.Element
    else:
        ParentClass = input_parent_class
    class temp_node(ParentClass):
        
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if classes is not None:
                if 'classes' in self:
                    self['classes'] = self['classes'] + classes
                else:
                    self['classes'] = classes

    temp_node.__name__ = f'{node_name}_node'
    return temp_node

def add_role_and_directive(app, node_type, name: str):
    app.add_directive(name, get_generic_directive(node_type))
    app.add_role(name, create_role(node_type))

class BaseNode(nodes.Element):
    pass

class BaseClassNode(BaseNode):
    classes = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)        
        if 'classes' in self:
            self['classes'] = self['classes'] + self.classes
        else:
            self['classes'] = self.classes

    def add_class(self, class_name: str):
        self['classes'] = self['classes'] + [class_name]

    def add_attributes(self, attribute_dict):
        a = 'attributes'
        if a in self:
            self[a].update(attribute_dict)
        else:
            self[a] = attribute_dict



from os import path

from .webslides_builder import WebslidesBuilder
from .slide import setup_slide
from .font_awesome import setup_fa
from .heading import setup_heading
from .header_footer import setup_header_footer
from .div import setup_div
from .link import setup_link
from .paragraph import setup_paragraph
from .span import setup_span
from .flexblock import setup_flexblock
from .hr import setup_hr
from .description_list import setup_description_list
from .preformatted import setup_preformatted
from .web import setup_web
from .video import setup_video
from .blockquote import setup_blockquote
from .template import setup_template

def setup(app):
    app.add_builder(WebslidesBuilder)
    app.add_html_theme('webslides_theme', path.join(path.abspath(path.dirname(__file__)), 'themes', 'webslides_base'))
    setup_slide(app)
    setup_fa(app)
    setup_header_footer(app)
    setup_div(app)
    setup_link(app)
    setup_paragraph(app)
    setup_span(app)
    setup_flexblock(app)
    setup_hr(app)
    setup_description_list(app)
    setup_preformatted(app)
    setup_heading(app)
    setup_web(app)
    setup_video(app)
    setup_blockquote(app)
    setup_template(app)


from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList
from sphinx.writers.html import HTMLTranslator
import jinja2
import yaml

environment = jinja2.Environment()

TD = 'template_definitions'

class template_node(nodes.Element): pass

class TemplateDefinitionDirective(SphinxDirective):
    required_arguments = 1
    has_content = True

    def run(self):
        try:
            len(self.env.app.config[TD])
        except AttributeError:
            self.env.app.config[TD] = {}
        self.env.app.config[TD][self.arguments[0]] = self.content.data
        return []

class TemplateExecuteDirective(SphinxDirective):
    required_arguments = 1
    optional_arguments = 1
    has_content = True
    option_spec = {
        'content_mode': directives.unchanged,
    }

    def run(self):
        
        rendered_content = []
        if 'content_mode' not in self.options:
            defined_input = '\n'.join(self.content.data)
            option_data = yaml.safe_load(defined_input)
            for template_line in self.env.app.config[TD][self.arguments[0]]:
                template = environment.from_string(template_line)
                rendered_content.append(template.render(**option_data))
        else:
            for template_line in self.env.app.config[TD][self.arguments[0]]:
                if '{{' in template_line and 'content' in template_line.lower() and '}}' in template_line:
                    char_index = 0
                    while template_line[char_index].isspace():
                        char_index +=1
                    white_space = template_line[0:char_index]
                    for line in self.content.data:
                        rendered_content.append(white_space+line)
                else:
                    rendered_content.append(template_line)

        return [self._process_content(rendered_content)]
        

    def _process_content(self, content):        
        par = template_node()
        self.state.nested_parse(
            ViewList(content, 'populated_template'), 
            self.content_offset, par
        )
        return par

def clear_templates(app, *args):
    """ I want to clear templates for each file.  No persistence"""
    if TD in app.config:
        app.config[TD] = {}

def setup_template(app):
    app.add_directive('template-define', TemplateDefinitionDirective)
    app.add_directive('template', TemplateExecuteDirective)
    app.connect('source-read', clear_templates)

class TemplateTranslator(HTMLTranslator):
    
    def visit_template_node(self, node):
        pass

    def depart_template_node(self, node):
        pass